#!/usr/bin/env python3
############################################################
## Jose F. Sanchez                                        ##
## Copyright (C) 2019-2020 Lauro Sumoy Lab, IGTP, Spain   ##
############################################################
## useful imports
import time
import io
import os
import re
import sys
from sys import argv
import subprocess
import traceback

from HCGB.functions import system_call_functions
from HCGB.functions import files_functions

############################################################
def create_genomeDir(folder, STAR_exe, num_threads, fasta_file, limitGenomeGenerateRAM):
    """Create the STAR_index genome dir 
    
    :param folder: folder to store the results
    :param STAR_exe: Executable path for STAR binary
    :param num_threads: number of threads to do the computation
    :param fasta_file: path to the genome directory
    :param limitGenomeGenerateRAM: limit RAM bytes to be used in the computation

    :type folder: string
    :type num_threads: int 
    :type STAR_exe: string
    :type fasta_file: string
    :type limitGenomeGenerateRAM: int

    :returns: genomeDir
    """
    ##
    genomeDir = files_functions.create_subfolder("STAR_index", folder)
    
    cmd_create = "%s --runMode genomeGenerate --limitGenomeGenerateRAM %s --runThreadN %s --genomeDir %s --genomeFastaFiles %s" %(
        STAR_exe, limitGenomeGenerateRAM, num_threads, genomeDir, fasta_file)

    print ('\t+ genomeDir generation for STAR mapping')
    create_code = system_call_functions.system_call(cmd_create, False, True)
    
    if not create_code:
        print ("** ERROR: Some error occurred during genomeDir creation... **")
        exit()
    
    return (genomeDir)

############################################################
def load_Genome(folder, STAR_exe, genomeDir, num_threads):
    """Load the genome stored in STAR_index  
    
    :param folder: folder where the genome will be loaded
    :param STAR_exe: Executable path for STAR binary
    :param num_threads: number of threads to do the computation
    :param genomeDir: path to the genome directory

    :type folder: string
    :type num_threads: int 
    :type STAR_exe: string
    :type genomeDir: string

    :returns: load_code
    """
    ## --genomeLoad LoadAndExit
    Load_folder = files_functions.create_subfolder('LoadMem', folder)
    cmd_LD = "%s --genomeDir %s --runThreadN %s --outFileNamePrefix %s --genomeLoad LoadAndExit" %(
        STAR_exe, genomeDir, num_threads, Load_folder)
    
    print ('\t+ Loading memory for STAR mapping')
    load_code = system_call_functions.system_call(cmd_LD, False, True)
    return (load_code)

############################################################
def remove_Genome(STAR_exe, genomeDir, folder, num_threads):
    """Remove the loaded genome 
    
    :param folder: folder where the genome was loaded
    :param STAR_exe: Executable path for STAR binary
    :param num_threads: number of threads to do the computation
    :param genomeDir: path to the genome directory

    :type folder: string
    :type num_threads: int 
    :type STAR_exe: string
    :type genomeDir: string

    :returns: remove_code
    """
    ## --genomeLoad Remove
    remove_folder = files_functions.create_subfolder('RemoveMem', folder)
    cmd_RM = "%s --genomeDir %s --outFileNamePrefix %s --runThreadN %s --genomeLoad Remove" %(
        STAR_exe, genomeDir, remove_folder, num_threads)
    
    ## send command    
    print ('\t+ Removing memory loaded for STAR mapping')
    remove_code = system_call_functions.system_call(cmd_RM, False, True)
    return (remove_code)

############################################################
def mapReads(option, reads, folder, name, STAR_exe, genomeDir, limitRAM_option, num_threads, Debug, multimapping):
    """
    Map reads using STAR software. Some parameters are set for small RNA Seq.

    Parameters set according to ENCODE Project directives for small RNAs
    https://www.encodeproject.org/rna-seq/small-rnas/
    
    :param option: If multiple files to map, use loaded genome (LoadAndKeep) if only one map, anything else.
    :param reads: List containing absolute path to reads (SE or PE)
    :param folder: Path for output results
    :param name: Sample name
    :param STAR_exe: Executable path for STAR binary
    :param genomeDir: path to the genome directory
    :param limitRAM_option: maximum available RAM (bytes) for map reads process. Default: 40000000000
    :param num_threads: number of threads to do the computation
    :param multimapping: Flag to say whether to use multimapping reads or not

    :type option: string
    :type reads: list
    :type folder: string 
    :type name: string 
    :type STAR_exe: string
    :type genomeDir: string 
    :type limitRAM_option: int
    :type num_threads: int
    :type multimapping: boolean
    
    :returns: mapping_code
    """
    ## open file
    print("\t+ Mapping sample %s using STAR" %name)
    
    if not os.path.isdir(folder):
        folder = files_functions.create_folder(folder)
    ##
    bam_file_name = os.path.join(folder, 'Aligned.sortedByCoord.out.bam')
        
    ## read is a list with 1 or 2 read fastq files
    jread = " ".join(reads)

    ## prepare command
    cmd = "%s --genomeDir %s --runThreadN %s " %(STAR_exe, genomeDir, num_threads)
    cmd = cmd + "--limitBAMsortRAM %s --outFileNamePrefix %s " %(limitRAM_option, folder + '/')

    if multimapping:
        cmd = cmd + " --outFilterMultimapNmax 10 "
    else:
        cmd = cmd + " --outFilterMultimapNmax 1 "

    ## some common options
    cmd = cmd + "--alignSJDBoverhangMin 1000 --outFilterMismatchNoverLmax 0.03 "
    cmd = cmd + "--outFilterScoreMinOverLread 0 --outFilterMatchNminOverLread 0 --outFilterMatchNmin 16 "
    cmd = cmd + "--alignIntronMax 1 --outSAMheaderHD @HD VN:1.4 SO:coordinate --outSAMtype BAM SortedByCoordinate "
    
    ## Multiple samples or just one?
    if option == 'LoadAndKeep':
        cmd = cmd + "--genomeLoad LoadAndKeep"
    else:
        cmd = cmd + "--genomeLoad NoSharedMemory"
    
    ## ReadFiles
    cmd = cmd + " --readFilesIn %s " %jread

    ## logfile & errfile
    logfile = os.path.join(folder, 'STAR.log')
    errfile = os.path.join(folder, 'STAR.err')
    cmd = cmd + ' > ' + logfile + ' 2> ' + errfile
    
    ## sent command
    mapping_code = system_call_functions.system_call(cmd, False, True)
    return (mapping_code)

###############

###########
def main():
    
    
    return 

######
if __name__== "__main__":
    main()
    
    
#######
# Usage: STAR  [options]... --genomeDir /path/to/genome/index/   --readFilesIn R1.fq R2.fq
# Spliced Transcripts Alignment to a Reference (c) Alexander Dobin, 2009-2019
########

#######
# For more details see:
# <https://github.com/alexdobin/STAR>
# <https://github.com/alexdobin/STAR/blob/master/doc/STARmanual.pdf>
# 
# ### versions
# versionGenome           2.7.1a
#     string: earliest genome index version compatible with this STAR release. Please do not change this value!
# 
# ### Parameter Files
# parametersFiles          -
#     string: name of a user-defined parameters file, "-": none. Can only be defined on the command line.
# 
# ### System
# sysShell            -
#     string: path to the shell binary, preferably bash, e.g. /bin/bash.
#                     - ... the default shell is executed, typically /bin/sh. This was reported to fail on some Ubuntu systems - then you need to specify path to bash.
# 
# ### Run Parameters
# runMode                         alignReads
#     string: type of the run.
# 
#                                 alignReads             ... map reads
#                                 genomeGenerate         ... generate genome files
#                                 inputAlignmentsFromBAM ... input alignments from BAM. Presently only works with --outWigType and --bamRemoveDuplicates.
#                                 liftOver               ... lift-over of GTF files (--sjdbGTFfile) between genome assemblies using chain file(s) from --genomeChainFiles.
# 
# runThreadN                      1
#     int: number of threads to run STAR
# 
# runDirPerm                      User_RWX
#     string: permissions for the directories created at the run-time.
#                                 User_RWX ... user-read/write/execute
#                                 All_RWX  ... all-read/write/execute (same as chmod 777)
# 
# runRNGseed                      777
#     int: random number generator seed.
# 
# 
# ### Genome Parameters
# genomeDir                   ./GenomeDir/
#     string: path to the directory where genome files are stored (for --runMode alignReads) or will be generated (for --runMode generateGenome)
# 
# genomeLoad                NoSharedMemory
#     string: mode of shared memory usage for the genome files. Only used with --runMode alignReads.
#                           LoadAndKeep     ... load genome into shared and keep it in memory after run
#                           LoadAndRemove   ... load genome into shared but remove it after run
#                           LoadAndExit     ... load genome into shared memory and exit, keeping the genome in memory for future runs
#                           Remove          ... do not map anything, just remove loaded genome from memory
#                           NoSharedMemory  ... do not use shared memory, each job will have its own private copy of the genome
# 
# genomeFastaFiles            -
#     string(s): path(s) to the fasta files with the genome sequences, separated by spaces. These files should be plain text FASTA files, they *cannot* be zipped.
#                             Required for the genome generation (--runMode genomeGenerate). Can also be used in the mapping (--runMode alignReads) to add extra (new) sequences to the genome (e.g. spike-ins).
# 
# genomeChainFiles            -
#     string: chain files for genomic liftover. Only used with --runMode liftOver .
# 
# genomeFileSizes             0
#     uint(s)>0: genome files exact sizes in bytes. Typically, this should not be defined by the user.
# 
# genomeConsensusFile         -
#     string: VCF file with consensus SNPs (i.e. alternative allele is the major (AF>0.5) allele)
# 
# ### Genome Indexing Parameters - only used with --runMode genomeGenerate
# genomeChrBinNbits           18
#     int: =log2(chrBin), where chrBin is the size of the bins for genome storage: each chromosome will occupy an integer number of bins. For a genome with large number of contigs, it is recommended to scale this parameter as min(18, log2[max(GenomeLength/NumberOfReferences,ReadLength)]).
# 
# genomeSAindexNbases         14
#     int: length (bases) of the SA pre-indexing string. Typically between 10 and 15. Longer strings will use much more memory, but allow faster searches. For small genomes, the parameter --genomeSAindexNbases must be scaled down to min(14, log2(GenomeLength)/2 - 1).
# 
# genomeSAsparseD             1
#     int>0: suffux array sparsity, i.e. distance between indices: use bigger numbers to decrease needed RAM at the cost of mapping speed reduction
# 
# genomeSuffixLengthMax       -1
#     int: maximum length of the suffixes, has to be longer than read length. -1 = infinite.
# 
# 
# ### Splice Junctions Database
# sjdbFileChrStartEnd                     -
#     string(s): path to the files with genomic coordinates (chr <tab> start <tab> end <tab> strand) for the splice junction introns. Multiple files can be supplied wand will be concatenated.
# 
# sjdbGTFfile                             -
#     string: path to the GTF file with annotations
# 
# sjdbGTFchrPrefix                        -
#     string: prefix for chromosome names in a GTF file (e.g. 'chr' for using ENSMEBL annotations with UCSC genomes)
# 
# sjdbGTFfeatureExon                      exon
#     string: feature type in GTF file to be used as exons for building transcripts
# 
# sjdbGTFtagExonParentTranscript          transcript_id
#     string: GTF attribute name for parent transcript ID (default "transcript_id" works for GTF files)
# 
# sjdbGTFtagExonParentGene                gene_id
#     string: GTF attribute name for parent gene ID (default "gene_id" works for GTF files)
# 
# sjdbGTFtagExonParentGeneName            gene_name
#     string(s): GTF attrbute name for parent gene name
# 
# sjdbGTFtagExonParentGeneType            gene_type gene_biotype
#     string(s): GTF attrbute name for parent gene type
# 
# sjdbOverhang                            100
#     int>0: length of the donor/acceptor sequence on each side of the junctions, ideally = (mate_length - 1)
# 
# sjdbScore                               2
#     int: extra alignment score for alignmets that cross database junctions
# 
# sjdbInsertSave                          Basic
#     string: which files to save when sjdb junctions are inserted on the fly at the mapping step
#                     Basic ... only small junction / transcript files
#                     All   ... all files including big Genome, SA and SAindex - this will create a complete genome directory
# 
# ### Variation parameters
# varVCFfile                              -
#     string: path to the VCF file that contains variation data.
# 
# ### Input Files
# inputBAMfile                -
#     string: path to BAM input file, to be used with --runMode inputAlignmentsFromBAM
# 
# ### Read Parameters
# readFilesType               Fastx
#     string: format of input read files
#                             Fastx       ... FASTA or FASTQ
#                             SAM SE      ... SAM or BAM single-end reads; for BAM use --readFilesCommand samtools view
#                             SAM PE      ... SAM or BAM paired-end reads; for BAM use --readFilesCommand samtools view
# 
# readFilesIn                 Read1 Read2
#     string(s): paths to files that contain input read1 (and, if needed,  read2)
# 
# readFilesPrefix             -
#     string: preifx for the read files names, i.e. it will be added in front of the strings in --readFilesIn
#                             -: no prefix
# 
# readFilesCommand             -
#     string(s): command line to execute for each of the input file. This command should generate FASTA or FASTQ text and send it to stdout
#                For example: zcat - to uncompress .gz files, bzcat - to uncompress .bz2 files, etc.
# 
# readMapNumber               -1
#     int: number of reads to map from the beginning of the file
#                             -1: map all reads
# 
# readMatesLengthsIn          NotEqual
#     string: Equal/NotEqual - lengths of names,sequences,qualities for both mates are the same  / not the same. NotEqual is safe in all situations.
# 
# readNameSeparator           /
#     string(s): character(s) separating the part of the read names that will be trimmed in output (read name after space is always trimmed)
# 
# readQualityScoreBase        33
#     int>=0: number to be subtracted from the ASCII code to get Phred quality score
# 
# clip3pNbases                 0
#     int(s): number(s) of bases to clip from 3p of each mate. If one value is given, it will be assumed the same for both mates.
# 
# clip5pNbases                 0
#     int(s): number(s) of bases to clip from 5p of each mate. If one value is given, it will be assumed the same for both mates.
# 
# clip3pAdapterSeq            -
#     string(s): adapter sequences to clip from 3p of each mate.  If one value is given, it will be assumed the same for both mates.
# 
# clip3pAdapterMMp            0.1
#     double(s): max proportion of mismatches for 3p adpater clipping for each mate.  If one value is given, it will be assumed the same for both mates.
# 
# clip3pAfterAdapterNbases    0
#     int(s): number of bases to clip from 3p of each mate after the adapter clipping. If one value is given, it will be assumed the same for both mates.
# 
# 
# ### Limits
# limitGenomeGenerateRAM               31000000000
#     int>0: maximum available RAM (bytes) for genome generation
# 
# limitIObufferSize                    150000000
#     int>0: max available buffers size (bytes) for input/output, per thread
# 
# limitOutSAMoneReadBytes              100000
#     int>0: max size of the SAM record (bytes) for one read. Recommended value: >(2*(LengthMate1+LengthMate2+100)*outFilterMultimapNmax
# 
# limitOutSJoneRead                    1000
#     int>0: max number of junctions for one read (including all multi-mappers)
# 
# limitOutSJcollapsed                  1000000
#     int>0: max number of collapsed junctions
# 
# limitBAMsortRAM                         0
#     int>=0: maximum available RAM (bytes) for sorting BAM. If =0, it will be set to the genome index size. 0 value can only be used with --genomeLoad NoSharedMemory option.
# 
# limitSjdbInsertNsj                     1000000
#     int>=0: maximum number of junction to be inserted to the genome on the fly at the mapping stage, including those from annotations and those detected in the 1st step of the 2-pass run
# 
# limitNreadsSoft                        -1
#     int: soft limit on the number of reads
# 
# ### Output: general
# outFileNamePrefix               ./
#     string: output files name prefix (including full or relative path). Can only be defined on the command line.
# 
# outTmpDir                       -
#     string: path to a directory that will be used as temporary by STAR. All contents of this directory will be removed!
#             - the temp directory will default to outFileNamePrefix_STARtmp
# 
# outTmpKeep                      None
#     string: whether to keep the tempporary files after STAR runs is finished
#                                 None ... remove all temporary files
#                                 All .. keep all files
# 
# outStd                          Log
#     string: which output will be directed to stdout (standard out)
#                                 Log                    ... log messages
#                                 SAM                    ... alignments in SAM format (which normally are output to Aligned.out.sam file), normal standard output will go into Log.std.out
#                                 BAM_Unsorted           ... alignments in BAM format, unsorted. Requires --outSAMtype BAM Unsorted
#                                 BAM_SortedByCoordinate ... alignments in BAM format, unsorted. Requires --outSAMtype BAM SortedByCoordinate
#                                 BAM_Quant              ... alignments to transcriptome in BAM format, unsorted. Requires --quantMode TranscriptomeSAM
# 
# outReadsUnmapped                None
#    string: output of unmapped and partially mapped (i.e. mapped only one mate of a paired end read) reads in separate file(s).
#                                 None    ... no output
#                                 Fastx   ... output in separate fasta/fastq files, Unmapped.out.mate1/2
# 
# outQSconversionAdd              0
#    int: add this number to the quality score (e.g. to convert from Illumina to Sanger, use -31)
# 
# outMultimapperOrder             Old_2.4
#     string: order of multimapping alignments in the output files
#                                 Old_2.4             ... quasi-random order used before 2.5.0
#                                 Random              ... random order of alignments for each multi-mapper. Read mates (pairs) are always adjacent, all alignment for each read stay together. This option will become default in the future releases.
# 
# ### Output: SAM and BAM
# outSAMtype                      SAM
#     strings: type of SAM/BAM output
#                                 1st word:
#                                 BAM  ... output BAM without sorting
#                                 SAM  ... output SAM without sorting
#                                 None ... no SAM/BAM output
#                                 2nd, 3rd:
#                                 Unsorted           ... standard unsorted
#                                 SortedByCoordinate ... sorted by coordinate. This option will allocate extra memory for sorting which can be specified by --limitBAMsortRAM.
# 
# outSAMmode                      Full
#     string: mode of SAM output
#                                 None ... no SAM output
#                                 Full ... full SAM output
#                                 NoQS ... full SAM but without quality scores
# 
# outSAMstrandField                               None
#     string: Cufflinks-like strand field flag
#                                 None        ... not used
#                                 intronMotif ... strand derived from the intron motif. Reads with inconsistent and/or non-canonical introns are filtered out.
# 
# outSAMattributes                Standard
#     string: a string of desired SAM attributes, in the order desired for the output SAM
#                                 NH HI AS nM NM MD jM jI XS MC ch ... any combination in any order
#                                 None        ... no attributes
#                                 Standard    ... NH HI AS nM
#                                 All         ... NH HI AS nM NM MD jM jI MC ch
#                                 vA          ... variant allele
#                                 vG          ... genomic coordiante of the variant overlapped by the read
#                                 vW          ... 0/1 - alignment does not pass / passes WASP filtering. Requires --waspOutputMode SAMtag
#                                 STARsolo:
#                                 CR CY UR UY ... sequences and quality scores of cell barcodes and UMIs for the solo* demultiplexing
#                                 CB UB       ... error-corrected cell barcodes and UMIs for solo* demultiplexing. Requires --outSAMtype BAM SortedByCoordinate.
#                                 sM          ... assessment of CB and UMI
#                                 sS          ... sequence of the entire barcode (CB,UMI,adapter...)
#                                 sQ          ... quality of the entire barcode
#                                 Unsupported/undocumented:
#                                 rB          ... alignment block read/genomic coordinates
#                                 vR          ... read coordinate of the variant
# 
# outSAMattrIHstart               1
#     int>=0:                     start value for the IH attribute. 0 may be required by some downstream software, such as Cufflinks or StringTie.
# 
# outSAMunmapped                  None
#     string(s): output of unmapped reads in the SAM format
#                                 1st word:
#                                 None   ... no output
#                                 Within ... output unmapped reads within the main SAM file (i.e. Aligned.out.sam)
#                                 2nd word:
#                                 KeepPairs ... record unmapped mate for each alignment, and, in case of unsorted output, keep it adjacent to its mapped mate. Only affects multi-mapping reads.
# 
# outSAMorder                     Paired
#     string: type of sorting for the SAM output
#                                 Paired: one mate after the other for all paired alignments
#                                 PairedKeepInputOrder: one mate after the other for all paired alignments, the order is kept the same as in the input FASTQ files
# 
# outSAMprimaryFlag        OneBestScore
#     string: which alignments are considered primary - all others will be marked with 0x100 bit in the FLAG
#                                 OneBestScore ... only one alignment with the best score is primary
#                                 AllBestScore ... all alignments with the best score are primary
# 
# outSAMreadID            Standard
#     string: read ID record type
#                                 Standard ... first word (until space) from the FASTx read ID line, removing /1,/2 from the end
#                                 Number   ... read number (index) in the FASTx file
# 
# outSAMmapqUnique        255
#     int: 0 to 255: the MAPQ value for unique mappers
# 
# outSAMflagOR           0
#     int: 0 to 65535: sam FLAG will be bitwise OR'd with this value, i.e. FLAG=FLAG | outSAMflagOR. This is applied after all flags have been set by STAR, and after outSAMflagAND. Can be used to set specific bits that are not set otherwise.
# 
# outSAMflagAND           65535
#     int: 0 to 65535: sam FLAG will be bitwise AND'd with this value, i.e. FLAG=FLAG & outSAMflagOR. This is applied after all flags have been set by STAR, but before outSAMflagOR. Can be used to unset specific bits that are not set otherwise.
# 
# outSAMattrRGline        -
#     string(s): SAM/BAM read group line. The first word contains the read group identifier and must start with "ID:", e.g. --outSAMattrRGline ID:xxx CN:yy "DS:z z z".
#             xxx will be added as RG tag to each output alignment. Any spaces in the tag values have to be double quoted.
#             Comma separated RG lines correspons to different (comma separated) input files in --readFilesIn. Commas have to be surrounded by spaces, e.g.
#             --outSAMattrRGline ID:xxx , ID:zzz "DS:z z" , ID:yyy DS:yyyy
# 
# outSAMheaderHD          -
#     strings: @HD (header) line of the SAM header
# 
# outSAMheaderPG          -
#     strings: extra @PG (software) line of the SAM header (in addition to STAR)
# 
# outSAMheaderCommentFile -
#     string: path to the file with @CO (comment) lines of the SAM header
# 
# outSAMfilter            None
#     string(s): filter the output into main SAM/BAM files
#                         KeepOnlyAddedReferences ... only keep the reads for which all alignments are to the extra reference sequences added with --genomeFastaFiles at the mapping stage.
#                         KeepAllAddedReferences ...  keep all alignments to the extra reference sequences added with --genomeFastaFiles at the mapping stage.
# 
# 
# outSAMmultNmax          -1
#     int: max number of multiple alignments for a read that will be output to the SAM/BAM files.
#                         -1 ... all alignments (up to --outFilterMultimapNmax) will be output
# 
# outSAMtlen              1
#     int: calculation method for the TLEN field in the SAM/BAM files
#                         1 ... leftmost base of the (+)strand mate to rightmost base of the (-)mate. (+)sign for the (+)strand mate
#                         2 ... leftmost base of any mate to rightmost base of any mate. (+)sign for the mate with the leftmost base. This is different from 1 for overlapping mates with protruding ends
# 
# outBAMcompression       1
#     int: -1 to 10  BAM compression level, -1=default compression (6?), 0=no compression, 10=maximum compression
# 
# outBAMsortingThreadN    0
#     int: >=0: number of threads for BAM sorting. 0 will default to min(6,--runThreadN).
# 
# outBAMsortingBinsN      50
#     int: >0:  number of genome bins fo coordinate-sorting
# 
# ### BAM processing
# bamRemoveDuplicatesType  -
#     string: mark duplicates in the BAM file, for now only works with (i) sorted BAM fed with inputBAMfile, and (ii) for paired-end alignments only
#                         -                       ... no duplicate removal/marking
#                         UniqueIdentical         ... mark all multimappers, and duplicate unique mappers. The coordinates, FLAG, CIGAR must be identical
#                         UniqueIdenticalNotMulti  ... mark duplicate unique mappers but not multimappers.
# 
# bamRemoveDuplicatesMate2basesN   0
#     int>0: number of bases from the 5' of mate 2 to use in collapsing (e.g. for RAMPAGE)
# 
# ### Output Wiggle
# outWigType          None
#     string(s): type of signal output, e.g. "bedGraph" OR "bedGraph read1_5p". Requires sorted BAM: --outSAMtype BAM SortedByCoordinate .
#                     1st word:
#                     None       ... no signal output
#                     bedGraph   ... bedGraph format
#                     wiggle     ... wiggle format
#                     2nd word:
#                     read1_5p   ... signal from only 5' of the 1st read, useful for CAGE/RAMPAGE etc
#                     read2      ... signal from only 2nd read
# 
# outWigStrand        Stranded
#     string: strandedness of wiggle/bedGraph output
#                     Stranded   ...  separate strands, str1 and str2
#                     Unstranded ...  collapsed strands
# 
# outWigReferencesPrefix    -
#     string: prefix matching reference names to include in the output wiggle file, e.g. "chr", default "-" - include all references
# 
# outWigNorm              RPM
#     string: type of normalization for the signal
#                         RPM    ... reads per million of mapped reads
#                         None   ... no normalization, "raw" counts
# 
# ### Output Filtering
# outFilterType                   Normal
#     string: type of filtering
#                                 Normal  ... standard filtering using only current alignment
#                                 BySJout ... keep only those reads that contain junctions that passed filtering into SJ.out.tab
# 
# outFilterMultimapScoreRange     1
#     int: the score range below the maximum score for multimapping alignments
# 
# outFilterMultimapNmax           10
#     int: maximum number of loci the read is allowed to map to. Alignments (all of them) will be output only if the read maps to no more loci than this value.
#          Otherwise no alignments will be output, and the read will be counted as "mapped to too many loci" in the Log.final.out .
# 
# outFilterMismatchNmax           10
#     int: alignment will be output only if it has no more mismatches than this value.
# 
# outFilterMismatchNoverLmax      0.3
#     real: alignment will be output only if its ratio of mismatches to *mapped* length is less than or equal to this value.
# 
# outFilterMismatchNoverReadLmax  1.0
#     real: alignment will be output only if its ratio of mismatches to *read* length is less than or equal to this value.
# 
# 
# outFilterScoreMin               0
#     int: alignment will be output only if its score is higher than or equal to this value.
# 
# outFilterScoreMinOverLread      0.66
#     real: same as outFilterScoreMin, but  normalized to read length (sum of mates' lengths for paired-end reads)
# 
# outFilterMatchNmin              0
#     int: alignment will be output only if the number of matched bases is higher than or equal to this value.
# 
# outFilterMatchNminOverLread     0.66
#     real: sam as outFilterMatchNmin, but normalized to the read length (sum of mates' lengths for paired-end reads).
# 
# outFilterIntronMotifs           None
#     string: filter alignment using their motifs
#                 None                           ... no filtering
#                 RemoveNoncanonical             ... filter out alignments that contain non-canonical junctions
#                 RemoveNoncanonicalUnannotated  ... filter out alignments that contain non-canonical unannotated junctions when using annotated splice junctions database. The annotated non-canonical junctions will be kept.
# 
# outFilterIntronStrands          RemoveInconsistentStrands
#     string: filter alignments
#                 RemoveInconsistentStrands      ... remove alignments that have junctions with inconsistent strands
#                 None                           ... no filtering
# 
# ### Output Filtering: Splice Junctions
# outSJfilterReads                All
#     string: which reads to consider for collapsed splice junctions output
#                 All: all reads, unique- and multi-mappers
#                 Unique: uniquely mapping reads only
# 
# outSJfilterOverhangMin          30  12  12  12
#     4 integers:    minimum overhang length for splice junctions on both sides for: (1) non-canonical motifs, (2) GT/AG and CT/AC motif, (3) GC/AG and CT/GC motif, (4) AT/AC and GT/AT motif. -1 means no output for that motif
#                                 does not apply to annotated junctions
# 
# outSJfilterCountUniqueMin       3   1   1   1
#     4 integers: minimum uniquely mapping read count per junction for: (1) non-canonical motifs, (2) GT/AG and CT/AC motif, (3) GC/AG and CT/GC motif, (4) AT/AC and GT/AT motif. -1 means no output for that motif
#                                 Junctions are output if one of outSJfilterCountUniqueMin OR outSJfilterCountTotalMin conditions are satisfied
#                                 does not apply to annotated junctions
# 
# outSJfilterCountTotalMin     3   1   1   1
#     4 integers: minimum total (multi-mapping+unique) read count per junction for: (1) non-canonical motifs, (2) GT/AG and CT/AC motif, (3) GC/AG and CT/GC motif, (4) AT/AC and GT/AT motif. -1 means no output for that motif
#                                 Junctions are output if one of outSJfilterCountUniqueMin OR outSJfilterCountTotalMin conditions are satisfied
#                                 does not apply to annotated junctions
# 
# outSJfilterDistToOtherSJmin     10  0   5   10
#     4 integers>=0: minimum allowed distance to other junctions' donor/acceptor
#                                 does not apply to annotated junctions
# 
# outSJfilterIntronMaxVsReadN        50000 100000 200000
#     N integers>=0: maximum gap allowed for junctions supported by 1,2,3,,,N reads
#                                 i.e. by default junctions supported by 1 read can have gaps <=50000b, by 2 reads: <=100000b, by 3 reads: <=200000. by >=4 reads any gap <=alignIntronMax
#                                 does not apply to annotated junctions
# 
# ### Scoring
# scoreGap                     0
#     int: splice junction penalty (independent on intron motif)
# 
# scoreGapNoncan               -8
#     int: non-canonical junction penalty (in addition to scoreGap)
# 
# scoreGapGCAG                 -4
#     GC/AG and CT/GC junction penalty (in addition to scoreGap)
# 
# scoreGapATAC                 -8
#     AT/AC  and GT/AT junction penalty  (in addition to scoreGap)
# 
# scoreGenomicLengthLog2scale   -0.25
#     extra score logarithmically scaled with genomic length of the alignment: scoreGenomicLengthLog2scale*log2(genomicLength)
# 
# scoreDelOpen                 -2
#     deletion open penalty
# 
# scoreDelBase                 -2
#     deletion extension penalty per base (in addition to scoreDelOpen)
# 
# scoreInsOpen                 -2
#     insertion open penalty
# 
# scoreInsBase                 -2
#     insertion extension penalty per base (in addition to scoreInsOpen)
# 
# scoreStitchSJshift           1
#     maximum score reduction while searching for SJ boundaries inthe stitching step
# 
# 
# ### Alignments and Seeding
# 
# seedSearchStartLmax             50
#     int>0: defines the search start point through the read - the read is split into pieces no longer than this value
# 
# seedSearchStartLmaxOverLread    1.0
#     real: seedSearchStartLmax normalized to read length (sum of mates' lengths for paired-end reads)
# 
# seedSearchLmax       0
#     int>=0: defines the maximum length of the seeds, if =0 max seed lengthis infinite
# 
# seedMultimapNmax      10000
#     int>0: only pieces that map fewer than this value are utilized in the stitching procedure
# 
# seedPerReadNmax       1000
#     int>0: max number of seeds per read
# 
# seedPerWindowNmax     50
#     int>0: max number of seeds per window
# 
# seedNoneLociPerWindow    10
#     int>0: max number of one seed loci per window
# 
# seedSplitMin                12
#     int>0: min length of the seed sequences split by Ns or mate gap
# 
# alignIntronMin              21
#     minimum intron size: genomic gap is considered intron if its length>=alignIntronMin, otherwise it is considered Deletion
# 
# alignIntronMax              0
#     maximum intron size, if 0, max intron size will be determined by (2^winBinNbits)*winAnchorDistNbins
# 
# alignMatesGapMax            0
#     maximum gap between two mates, if 0, max intron gap will be determined by (2^winBinNbits)*winAnchorDistNbins
# 
# alignSJoverhangMin          5
#     int>0: minimum overhang (i.e. block size) for spliced alignments
# 
# alignSJstitchMismatchNmax   0 -1 0 0
#     4*int>=0: maximum number of mismatches for stitching of the splice junctions (-1: no limit).
#                             (1) non-canonical motifs, (2) GT/AG and CT/AC motif, (3) GC/AG and CT/GC motif, (4) AT/AC and GT/AT motif.
# 
# alignSJDBoverhangMin        3
#     int>0: minimum overhang (i.e. block size) for annotated (sjdb) spliced alignments
# 
# alignSplicedMateMapLmin     0
#     int>0: minimum mapped length for a read mate that is spliced
# 
# alignSplicedMateMapLminOverLmate 0.66
#     real>0: alignSplicedMateMapLmin normalized to mate length
# 
# alignWindowsPerReadNmax     10000
#     int>0: max number of windows per read
# 
# alignTranscriptsPerWindowNmax       100
#     int>0: max number of transcripts per window
# 
# alignTranscriptsPerReadNmax               10000
#     int>0: max number of different alignments per read to consider
# 
# alignEndsType           Local
#     string: type of read ends alignment
#                         Local             ... standard local alignment with soft-clipping allowed
#                         EndToEnd          ... force end-to-end read alignment, do not soft-clip
#                         Extend5pOfRead1   ... fully extend only the 5p of the read1, all other ends: local alignment
#                         Extend5pOfReads12 ... fully extend only the 5p of the both read1 and read2, all other ends: local alignment
# 
# alignEndsProtrude       0    ConcordantPair
#     int, string:        allow protrusion of alignment ends, i.e. start (end) of the +strand mate downstream of the start (end) of the -strand mate
#                         1st word: int: maximum number of protrusion bases allowed
#                         2nd word: string:
#                                             ConcordantPair ... report alignments with non-zero protrusion as concordant pairs
#                                             DiscordantPair ... report alignments with non-zero protrusion as discordant pairs
# 
# alignSoftClipAtReferenceEnds    Yes
#     string: allow the soft-clipping of the alignments past the end of the chromosomes
#                                 Yes ... allow
#                                 No  ... prohibit, useful for compatibility with Cufflinks
# 
# alignInsertionFlush     None
#     string: how to flush ambiguous insertion positions
#                         None    ... insertions are not flushed
#                         Right   ... insertions are flushed to the right
# 
# ### Paired-End reads
# peOverlapNbasesMin          0
#     int>=0:             minimum number of overlap bases to trigger mates merging and realignment
# 
# peOverlapMMp                0.01
#     real, >=0 & <1:     maximum proportion of mismatched bases in the overlap area
# 
# ### Windows, Anchors, Binning
# 
# winAnchorMultimapNmax           50
#     int>0: max number of loci anchors are allowed to map to
# 
# winBinNbits                     16
#     int>0: =log2(winBin), where winBin is the size of the bin for the windows/clustering, each window will occupy an integer number of bins.
# 
# winAnchorDistNbins              9
#     int>0: max number of bins between two anchors that allows aggregation of anchors into one window
# 
# winFlankNbins                   4
#     int>0: log2(winFlank), where win Flank is the size of the left and right flanking regions for each window
# 
# winReadCoverageRelativeMin      0.5
#     real>=0: minimum relative coverage of the read sequence by the seeds in a window, for STARlong algorithm only.
# 
# winReadCoverageBasesMin      0
#     int>0: minimum number of bases covered by the seeds in a window , for STARlong algorithm only.
# 
# ### Chimeric Alignments
# chimOutType                 Junctions
#     string(s): type of chimeric output
#                             Junctions       ... Chimeric.out.junction
#                             SeparateSAMold  ... output old SAM into separate Chimeric.out.sam file
#                             WithinBAM       ... output into main aligned BAM files (Aligned.*.bam)
#                             WithinBAM HardClip  ... (default) hard-clipping in the CIGAR for supplemental chimeric alignments (defaultif no 2nd word is present)
#                             WithinBAM SoftClip  ... soft-clipping in the CIGAR for supplemental chimeric alignments
# 
# chimSegmentMin              0
#     int>=0: minimum length of chimeric segment length, if ==0, no chimeric output
# 
# chimScoreMin                0
#     int>=0: minimum total (summed) score of the chimeric segments
# 
# chimScoreDropMax            20
#     int>=0: max drop (difference) of chimeric score (the sum of scores of all chimeric segments) from the read length
# 
# chimScoreSeparation         10
#     int>=0: minimum difference (separation) between the best chimeric score and the next one
# 
# chimScoreJunctionNonGTAG    -1
#     int: penalty for a non-GT/AG chimeric junction
# 
# chimJunctionOverhangMin     20
#     int>=0: minimum overhang for a chimeric junction
# 
# chimSegmentReadGapMax       0
#     int>=0: maximum gap in the read sequence between chimeric segments
# 
# chimFilter                  banGenomicN
#     string(s): different filters for chimeric alignments
#                             None ... no filtering
#                             banGenomicN ... Ns are not allowed in the genome sequence around the chimeric junction
# 
# chimMainSegmentMultNmax        10
#     int>=1: maximum number of multi-alignments for the main chimeric segment. =1 will prohibit multimapping main segments.
# 
# chimMultimapNmax                    0
#     int>=0: maximum number of chimeric multi-alignments
#                                 0 ... use the old scheme for chimeric detection which only considered unique alignments
# 
# chimMultimapScoreRange          1
#     int>=0: the score range for multi-mapping chimeras below the best chimeric score. Only works with --chimMultimapNmax > 1
# 
# chimNonchimScoreDropMin         20
#     int>=0: to trigger chimeric detection, the drop in the best non-chimeric alignment score with respect to the read length has to be greater than this value
# 
# chimOutJunctionFormat           0
#     int: formatting type for the Chimeric.out.junction file
#                                 0 ... no comment lines/headers
#                                 1 ... comment lines at the end of the file: command line and Nreads: total, unique, multi
# 
# ### Quantification of Annotations
# quantMode                   -
#     string(s): types of quantification requested
#                             -                ... none
#                             TranscriptomeSAM ... output SAM/BAM alignments to transcriptome into a separate file
#                             GeneCounts       ... count reads per gene
# 
# quantTranscriptomeBAMcompression    1       1
#     int: -2 to 10  transcriptome BAM compression level
#                             -2  ... no BAM output
#                             -1  ... default compression (6?)
#                              0  ... no compression
#                              10 ... maximum compression
# 
# quantTranscriptomeBan       IndelSoftclipSingleend
#     string: prohibit various alignment type
#                             IndelSoftclipSingleend  ... prohibit indels, soft clipping and single-end alignments - compatible with RSEM
#                             Singleend               ... prohibit single-end alignments
# 
# ### 2-pass Mapping
# twopassMode                 None
#     string: 2-pass mapping mode.
#                             None        ... 1-pass mapping
#                             Basic       ... basic 2-pass mapping, with all 1st pass junctions inserted into the genome indices on the fly
# 
# twopass1readsN              -1
#     int: number of reads to process for the 1st step. Use very large number (or default -1) to map all reads in the first step.
# 
# 
# ### WASP parameters
# waspOutputMode              None
#     string: WASP allele-specific output type. This is re-implemenation of the original WASP mappability filtering by Bryce van de Geijn, Graham McVicker, Yoav Gilad & Jonathan K Pritchard. Please cite the original WASP paper: Nature Methods 12, 1061–1063 (2015), https://www.nature.com/articles/nmeth.3582 .
#                             SAMtag      ... add WASP tags to the alignments that pass WASP filtering
# 
# ### STARsolo (single cell RNA-seq) parameters
# soloType                    None
#     string(s): type of single-cell RNA-seq
#                             CB_UMI_Simple   ... (a.k.a. Droplet) one UMI and one Cell Barcode of fixed length in read2, e.g. Drop-seq and 10X Chromium
#                             CB_UMI_Complex  ... one UMI of fixed length, but multiple Cell Barcodes of varying length, as well as adapters sequences are allowed in read2 only, e.g. inDrop.
# 
# soloCBwhitelist             -
#     string(s): file(s) with whitelist(s) of cell barcodes. Only one file allowed with 
# 
# soloCBstart                 1
#     int>0: cell barcode start base
# 
# soloCBlen                   16
#     int>0: cell barcode length
# 
# soloUMIstart                17
#     int>0: UMI start base
# 
# soloUMIlen                  10
#     int>0: UMI length
# 
# soloBarcodeReadLength       1
#     int: length of the barcode read
#                             1   ... equal to sum of soloCBlen+soloUMIlen
#                             0   ... not defined, do not check
# 
# soloCBposition              -
#     strings(s)              position of Cell Barcode(s) on the barcode read.
#                             Presently only works with --soloType CB_UMI_Complex, and barcodes are assumed to be on Read2.
#                             Format for each barcode: startAnchor_startDistance_endAnchor_endDistance
#                             start(end)Anchor defines the anchor base for the CB: 0: read start; 1: read end; 2: adapter start; 3: adapter end
#                             start(end)Distance is the distance from the CB start(end) to the Anchor base
#                             String for different barcodes are separated by space.
#                             Example: inDrop (Zilionis et al, Nat. Protocols, 2017):
#                             --soloCBposition  0_0_2_-1  3_1_3_8
# 
# soloUMIposition             -
#     string                  position of the UMI on the barcode read, same as soloCBposition
#                             Example: inDrop (Zilionis et al, Nat. Protocols, 2017):
#                             --soloCBposition  3_9_3_14
# 
# soloAdapterSequence         -
#     string:                 adapter sequence to anchor barcodes.
# 
# soloAdapterMismatchesNmax   1
#     int>0:                  maximum number of mismatches allowed in adapter sequence.
# 
# soloCBmatchWLtype           1MM_multi
#     string:                 matching the Cell Barcodes to the WhiteList
#                             Exact                   ... only exact matches allowed
#                             1MM                     ... only one match in whitelist with 1 mismatched base allowed. Allowed CBs have to have at least one read with exact match.
#                             1MM_multi               ... multiple matches in whitelist with 1 mismatched base allowed, posterior probability calculation is used choose one of the matches. 
#                                                         Allowed CBs have to have at least one read with exact match. Similar to CellRanger 2.2.0
#                             1MM_multi_pseudocounts  ... same as 1MM_Multi, but pseudocounts of 1 are added to all whitelist barcodes.
#                                                         Similar to CellRanger 3.x.x
# 
# soloStrand                  Forward
#     string: strandedness of the solo libraries:
#                             Unstranded  ... no strand information
#                             Forward     ... read strand same as the original RNA molecule
#                             Reverse     ... read strand opposite to the original RNA molecule
# 
# soloFeatures                Gene
#     string(s):              genomic features for which the UMI counts per Cell Barcode are collected
#                             Gene            ... genes: reads match the gene transcript
#                             SJ              ... splice junctions: reported in SJ.out.tab
#                             GeneFull        ... full genes: count all reads overlapping genes' exons and introns
#                             Transcript3p   ... quantification of transcript for 3' protocols
# 
# soloUMIdedup                1MM_All
#     string(s):              type of UMI deduplication (collapsing) algorithm
#                             1MM_All             ... all UMIs with 1 mismatch distance to each other are collapsed (i.e. counted once)
#                             1MM_Directional     ... follows the "directional" method from the UMI-tools by Smith, Heger and Sudbery (Genome Research 2017).
#                             Exact               ... only exactly matching UMIs are collapsed
# 
# soloUMIfiltering            -
#     string(s)               type of UMI filtering
#                             -               ... basic filtering: remove UMIs with N and homopolymers (similar to CellRanger 2.2.0)
#                             MultiGeneUMI    ... remove lower-count UMIs that map to more than one gene (introduced in CellRanger 3.x.x)
# 
# soloOutFileNames            Solo.out/          features.tsv barcodes.tsv        matrix.mtx
#     string(s)               file names for STARsolo output:
#                             file_name_prefix   gene_names   barcode_sequences   cell_feature_count_matrix
# 
# soloCellFilter              CellRanger2.2 3000 0.99 10
#     string(s):              cell filtering type and parameters
#                             CellRanger2.2   ... simple filtering of CellRanger 2.2, followed by thre numbers: number of expected cells, robust maximum percentile for UMI count, maximum to minimum ratio for UMI count
#                             TopCells        ... only report top cells by UMI count, followed by the excat number of cells
#                             None            ... do not output filtered cells
# 
#######



#!/usr/bin/env python3
############################################################
## Jose F. Sanchez                                        ##
## Copyright (C) 2019-2020 Lauro Sumoy Lab, IGTP, Spain   ##
############################################################
## useful imports
import time
import io
import os
import re
import sys
from sys import argv
import subprocess
import argparse
from termcolor import colored
import pandas as pd

from XICRA.config import set_config
from XICRA.scripts import bedtools_caller

import HCGB.functions.aesthetics_functions as HCGB_aes
import HCGB.functions.files_functions as HCGB_files
import HCGB.functions.system_call_functions as HCGB_sys
import HCGB.functions.time_functions as HCGB_time
import HCGB.format_conversion

#######################################################
def parse_bed_length(bed_file, sample, debug):

    #read a file
    fileReader = open(bed_file)
    
    header_list = [ "transcript", "biotype", "gene", "length", "chr", "count", "sample", "strand"]    
    print("\t".join(header_list))

    ## skip comments at the beginning of files
    while True:
        line = fileReader.readline()
        line = line.rstrip()
        field=line.strip().split('\t')

        ## example
        # 1    1218699    1218706    1    1    1217688    1218768    ENST00000494748    100    -    1217688    1218768    255,0,0    3    0,311,1837    2583,176,757    ENSG00000078808    retained_intron

        if line:
            
            ## ATTENTION: It takes many time and RAM to store results.
            ## Dump into file directly
            ## save data
            #data_parse.loc[field[7], 'sample'] = sample
            #data_parse.loc[field[7],'biotype'] = field[17]
            #data_parse.loc[field[7],'gene'] = field[16]
            #data_parse.loc[field[7],'length'] = abs(int(field[2])-int(field[1]))
            #data_parse.loc[field[7],'chr'] = field[0]
            #data_parse.loc[field[7],'count'] = field[3]
            #data_parse.loc[field[7],'strand'] = field[9]
            
            length_match = abs(int(field[2])-int(field[1]))
            
            ##          transcript  biotype    gene        length                           chr       count     sample  strand    
            list2print = [ field[7], field[17], field[16], str(length_match), str(field[0]), str(field[3]), sample, field[9] ]
            
            print("\t".join(list2print))
        else:
            fileReader.close()
            break

#######################################################
def get_length_dist(bam_file, GTF_info, name, chr, path_given, folder_GTF, debug=False):
    """
    Gets read length distribution from mapping file
    
    It initially converts BAM -> BED format, then checks if annotation file provided is splitted 
    and converted to BED format.
    
    It then intersects annotations provided and mapping reads and generates a table with statistics.

    :param bam_file: Mapping file in BAM format
    :param GTF_info: GTF annotation file
    :param name: Name of the sample
    :param chr: Option that states if Chr splitted is already done and GTF_info is a subset file
    :param path_given: Path to store results
    :param folder_GTF: Path to store splitted GTF files
    :param debug: Wether to show debug messages or not.
    
    :type bam_file: string
    :type GTF_info: string
    :type name: string
    :type chr: boolean
    :type path_given: string
    :type folder_GTF: string
    :type debug: boolean
    """
    
    ## debugging messages
    if debug:
        HCGB_aes.debug_message("******************")
        HCGB_aes.debug_message("get_length_dist function call: ", color="yellow")
        print()
        HCGB_aes.debug_message("bam_file: " + bam_file, color="yellow")
        HCGB_aes.debug_message("GTF_info: " + GTF_info, color="yellow")
        HCGB_aes.debug_message("name: " + name, color="yellow")
        HCGB_aes.debug_message("path_given: " + path_given, color="yellow")
        HCGB_aes.debug_message("folder_GTF: " + folder_GTF, color="yellow")
        HCGB_aes.debug_message("chr: " + str(chr), color="yellow")
        
    
    print("+ Converting BAM file to BED...")    
    
    ## converts BAM file in BED format
    mapping_bed_file = bedtools_caller.convert_bam2bed(name, bam_file, path_given, debug=debug)
    
    ## convert annotation in GTF to BED
    print("+ Converting GTF annotation file into BED...")
    if chr: 
        ## GTF_info is already splitted:
        (bed_files, gtf_files) = convert_GTF2bed(GTF_info, folder_GTF, num_files=0, debug=debug)
    else:    
        (bed_files, gtf_files) = convert_GTF2bed(GTF_info, folder_GTF, debug=debug)

    ## debugging messages
    if debug:
        HCGB_aes.debug_message("bed_files", color="yellow")
        print(bed_files)        
        print()
        HCGB_aes.debug_message("gtf_files", color="yellow")
        print(gtf_files)
        
    ## Loop for each reference sequence BED file
    ## and intersect with mapping BAM->BED file
    
    print("+ Intersecting GTF annotation file and mapping BED file...")

    original_stdout = sys.stdout # Save a reference to the original standard output

    for RefSeq_bed_file in bed_files.values():
        ## intersect
        ## call bedtools_caller.intersect_coordinates()
        bed_file = bedtools_caller.intersect_coordinates(mapping_bed_file, RefSeq_bed_file, path_given, name, "-s", debug)
        out_file = os.path.join(path_given, name + ".length.results.txt") 
    
        ## print results to file
        with open(out_file, 'w') as f:
            sys.stdout = f # Change the standard output to the file we created.
            
            ## print results directly to open file
            parse_bed_length(bed_file, name, debug)
            f.close()
    
    ## Reset print standard out
    sys.stdout = original_stdout
    
    ## Maybe merge all data from different chromosomes
    
    ## call R script in XICRA.stats to plot results
    
    exit()
    
#######################################################
def convert_GTF2bed(GTF_file, path_given, cpus2use=2, num_files=1, debug=False):
    """
    This functions calls gtf2bed from HCGB to generate a conversion from GTF to BED format.
    
    It allows to split big GTF file provided into multiple files and convert each subfile into BED format.
    
    """
    
    HCGB_files.create_folder(path_given)
    
    split_GTF_files = {}
    if num_files==0:
        ## GTF file is already splitted
        split_GTF_files['GTF_file'] = os.path.abspath(GTF_file)
    else:
        
        print("+ Split GTF in multiple subsets to speed up process")
                
        ## split GTF file
        split_GTF_files = HCGB.format_conversion.file_splitter.split_file_call(GTF_file, num_files=1, 
                                                 name=False, chr_option=True, 
                                                 in_format="GTF",
                                                 path_given=path_given, debug=debug) 
    # create bed
    print("+ Converting GTF files in BED format:")
    bed_files = {}
    for each_file in split_GTF_files.values():
        ## create name
        each_file_name = HCGB_files.get_file_name(each_file)
        bed_file = os.path.join(path_given, each_file_name + ".bed") ## create a name
        
        print("\t- Converting GTF file: " + each_file)
        print("\t- To BED file: " + each_file)
        print()
        ## convert GTF
        HCGB.format_conversion.gtf2bed.parse_GTF_call(each_file, bed_file, debug)
        
        ## save
        bed_files[each_file] = bed_file
    
    return (bed_files, split_GTF_files)

#######################################################
def main():
    ## this code runs when call as a single script
    parser=argparse.ArgumentParser(description='''Get length distribution for biotypes mapped in BAM file
    Note:
    It takes into account no genes or transcript are broken
    It is also possible to split according to chromosome (one gtf/chromosome)
    
    ''');
    
    parser.add_argument('--input', '-i', help='Input BAM file', required=True);

    parser.add_argument('--name', '-n',
                        help='Name of the sample. Default: use filename provided.', default="");
    
    parser.add_argument('--path', '-p',
                        help='Path to save results generated. Default: use path from filename provided.', default="");

    parser.add_argument('--path_GTF',
                        help='Path to store results generated from split GTF file. Default: use path provided previously.', default="");

    parser.add_argument('--annot', '-a',
                        help='Absolute path for GTF file.', default="");

    parser.add_argument('--annot_splitted', '-as', action="store_true",
                        help='--annot is a dir containing splitted GTF files for each chromosome.');
    
    args=parser.parse_args();
    
    ## lets split the big file provided
    print()
    print("** get_length_distribution script call **")
    print()
    files_generated = get_length_dist(os.path.abspath(args.input), os.path.abspath(args.annot), name=args.name, 
              chr=args.annot_splitted, path_given=os.path.abspath(args.path), folder_GTF=os.path.abspath(args.path_GTF), 
              debug=False)
    return ()

######
if __name__== "__main__":
    main()
    


#!/usr/bin/env python3

"""
test example file
"""

import os
import HCGB.functions
import shutil

## Download from git subset2test folder and run sh command

def run_test(options):
    print ("** Downloading folder containing test reads from github")
    
    ## Download repo
    str_command = "gitdir https://github.com/HCGB-IGTP/XICRA/tree/master/subset2test/"
    HCGB.functions.system_call_functions.system_call(str_command)

    shutil.move("./subset2test/subset_PE/", "./subset_PE/")
    shutil.move("./subset2test/subset_SE/", "./subset_SE/")
    shutil.move("./subset2test/subset_tRNA/", "./subset_tRNA/")
    shutil.move("./subset2test/test_subset.sh", "./test_subset.sh")
    
    shutil.rmtree("./subset2test/")
    
    print ("** Now, either execute sh test_subset.sh or step by step")
    



#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez, Marta Lopez & Lauro Sumoy           ##
## Copyright (C) 2019-2021 Lauro Sumoy Lab, IGTP, Spain ##
##########################################################
'''
Calls optimir to create miRNA profile
'''
## useful imports
import time
import io
import os
import re
import sys
from sys import argv
from io import open
from termcolor import colored

## import my modules
from HCGB import functions
from XICRA.config import set_config

###############       
def optimir (reads, outpath, file_name, num_threads, matureFasta, hairpinFasta, miRNA_gff, Debug):
    """Executes the optimiR analyis of the sample.

    Checks if the reads are joined and builds the optimiR command to execute it. 
    
    :param reads: file with sample reads
    :param outpath: output folder
    :param file_name: sample name
    :param num_threads: selected threads (by defoult 2)
    :param matureFasta: mature fasta file 
    :param hairpinFasta: hairpin fasta file
    :param miRNA_gff: miRNA gff3 annotation file
    :param Debug: display complete log.


    :returns: The optimiR output
    """
    optimir_exe = set_config.get_exe("optimir", Debug=Debug)
    sRNAbench_db = os.path.abspath(os.path.join(os.path.dirname(optimir_exe), '..')) ## optimir
    logfile = os.path.join(outpath, 'optimir.log')
    errfile = os.path.join(outpath, 'optimir.err')
    
    if (len(reads) > 1):
        print (colored("** ERROR: Only 1 fastq file is allowed please joined reads before...", 'red'))
        exit()
 
    ## create command  
    cmd = "%s process --fq %s --gff_out -o %s --maturesFasta %s --hairpinsFasta %s --gff3 %s > %s 2> %s" %(
        optimir_exe, reads[0], outpath, matureFasta, hairpinFasta, miRNA_gff, logfile, errfile)
    return(functions.system_call_functions.system_call(cmd))


###############       
def optimir_caller(reads, sample_folder, name, threads, matureFasta, hairpinFasta, miRNA_gff, species, Debug):
    """Passes the sample information to be analyzed by optimiR().

    Checks if the computation has been performed before. If not, it calls
    optimir().
    
    :param reads: file with sample reads
    :param sample_folder: output folder
    :param name: sample name
    :param threads: selected threads (by defoult 2)
    :param matureFasta: mature fasta file 
    :param hairpinFasta: hairpin fasta file
    :param miRNA_gff: miRNA gff3 annotation file
    :param species: species tag ID. Default: hsa (Homo sapiens)
    :param Debug: display complete log.


    :returns: True/False
    """
    # check if previously generated and succeeded
    filename_stamp = sample_folder + '/.success'
    if os.path.isfile(filename_stamp):
        stamp = functions.time_functions.read_time_stamp(filename_stamp)
        print (colored("\tA previous command generated results on: %s [%s -- %s]" %(stamp, name, 'OptimiR'), 'yellow'))
    else:
        # Call sRNAbench
    ## no species option for OptimiR
        code_returned = optimir(reads, sample_folder, name, threads, matureFasta, hairpinFasta, miRNA_gff,  Debug)
        if code_returned:
            functions.time_functions.print_time_stamp(filename_stamp)
        else:
            print ('** Sample %s failed...' %name)
            return(False)
        
    return(True)



#usr/bin/env python

## useful imports
import time
import io
import os
import re
import sys
import csv
from sys import argv
import subprocess
from collections import defaultdict
import argparse
from termcolor import colored
import concurrent.futures

from XICRA.scripts import bedtools_caller
from XICRA.scripts import samtools_caller
from XICRA.scripts import pilfer_caller
from XICRA.config import set_config
from XICRA.modules import database

import HCGB.functions.fasta_functions as HCGB_fasta
import HCGB.functions.time_functions as HCGB_time
import HCGB.functions.system_call_functions as HCGB_sys
import HCGB.functions.files_functions as HCGB_files
import HCGB.functions.main_functions as HCGB_main
import HCGB.functions.aesthetics_functions as HCGB_aes
import HCGB.format_conversion.file_splitter as HCGB_splitter

##########################################################3
def annotate_sam_call(sam_file, gold_piRNA, ncpu, folder, Debug):
    
    ### Original code
    # def ReverseComplement(seq):
    #    seq_dict = {'A':'T','T':'A','G':'C','C':'G','N':'N'}
    #    return "".join([seq_dict[base] for base in reversed(seq)])
    #
    # seq_id = []
    # with open(sys.argv[1],"rb") as listin:
    #    for seq in listin:
    #        seq_id.append(seq.strip())
    # seq_id = list(set(seq_id))
    
    
    ## read fasta file and save results in list
    seq_dict = HCGB_fasta.get_fasta_dict(gold_piRNA, Debug=Debug)
    seq_id = list(seq_dict.keys())

    if Debug:
        print("seq_dict gold piRNA")
        ##print(seq_dict)
        print ("ATTENTION: Very big file. See file provided for details: " + gold_piRNA)

    ## as it might be very big, we are splitting and processing in parallel
    path_given = HCGB_files.create_folder(os.path.join(folder, "split_sam"))
    HCGB_splitter.split_file_call(sam_file, ncpu*8, "split_file", False, 'SAM', os.path.abspath(path_given), Debug)
    
    list_sam_files = HCGB_main.get_fullpath_list(path_given, Debug)
    
    ## remove non-desired files
    list_sam_files = [s for s in list_sam_files if '.parsed' not in s]
    list_sam_files = [s for s in list_sam_files if '.split_file_success' not in s]

    ## send for each subset using multiple threads
    cpus2use=ncpu*4 ## Each single process uses just 10% each cpu, we would increase speed by using x4 times more. 
		    ## It might not be best way but it really speeds up the process and does not consum many RAM or CPUs
    with concurrent.futures.ThreadPoolExecutor(max_workers=cpus2use) as executor:
        commandsSent = { executor.submit(pilfer_caller.annotate_sam, seq_id, 
                                         subset_sam, Debug): subset_sam for subset_sam in list_sam_files }

        for cmd2 in concurrent.futures.as_completed(commandsSent):
            details = commandsSent[cmd2]
            try:
                data = cmd2.result()
            except Exception as exc:
                print ('***ERROR:')
                print (cmd2)
                print('%r generated an exception: %s' % (details, exc))

    list_sam_files = (s + '.parsed' for s in list_sam_files)
    
    if Debug:
        print(list_sam_files)

    return(list_sam_files)

################################
def process_call(bam_file, sample_folder, name, gold_piRNA, ncpu, Debug):
    
    ## check if previously trimmed and succeeded
    filename_stamp = sample_folder + '/.success_BAM2PILFER'
    if os.path.isfile(filename_stamp):
        print ("\n+ Converting BAM file into PILFER input file")
        stamp = HCGB_time.read_time_stamp(filename_stamp)
        print (colored("\tA previous command generated results on: %s [%s -- %s]" %(stamp, name, 'BAMtoPILFER'), 'yellow'))
        
        pilfer_file = os.path.join(sample_folder, name + ".pilfer.bed")
        return (pilfer_file)

    else:
        code_returned = bam2pilfer(bam_file, sample_folder, name, gold_piRNA, ncpu, Debug)
        if code_returned:
            HCGB_time.print_time_stamp(filename_stamp)
            return (code_returned) ## file name
        else:
            print ('** Sample %s failed...' %name)
            return(False)

################################
def merge_sam_bed(bed_file, sam_file, pilfer_tmp, Debug):

    ## TODO:
    ## Linux only: find an alternative
    ## paste Aligned.sortedByCoord.out.bed Aligned.sortedByCoord.out.sam | awk -v "OFS=\t" '{print $1, $2, $3, $16, $6}'

    if Debug:
        HCGB_aes.debug_message("bed_file: " + bed_file, "yellow")
        HCGB_aes.debug_message("sam_file: " + sam_file, "yellow")
        HCGB_aes.debug_message("pilfer_tmp: " + pilfer_tmp, "yellow")

    cmd_paste = "paste %s %s | grep \':P\' | awk -v \"OFS=\t\" \'{print $1, $2, $3, $16, $6}\' > %s" %(bed_file, sam_file, pilfer_tmp)
        
    paste_code = HCGB_sys.system_call(cmd_paste, False, True)
    return paste_code

################################
def bam2pilfer(bam_file, out_folder, name, annot_info, ncpu, Debug):
    
    out_folder = os.path.abspath(out_folder)
    
    print("+ Convert BAM to PILFER input file")
    print("+ Several pre-processing steps are required:")
    
    ## convert BAM2bed
    print("\t- Convert original BAM to BED format file...")
    bed_file = bedtools_caller.convert_bam2bed(name + '_original', bam_file, out_folder, pilfer=True, debug=Debug)
    
    ## substract ncRNA included (no piRNA)
    print("\t- Subtract ncRNA annotated reads from BAM to reduce processing...")
    bed_file_reduced = bedtools_caller.subtract_coordinates(bed_file, annot_info['general']['ncRNA'], 
                                                            out_folder, name + '_subtracted', "", Debug)
    
    ## get IDs and convert original bam to sam only with retained IDs
    ## required samtools 1.12
    print("\t- Get subtracted reads...")
    ## generate paste filter tmp file
    reads_to_get = os.path.join(out_folder, "reads2retain.txt")
    filename_stamp = out_folder + '/.subtract_read_success'
    if os.path.isfile(filename_stamp) and HCGB_files.is_non_zero_file(reads_to_get):
        stamp = HCGB_time.read_time_stamp(filename_stamp)
        print (colored("\tA previous command generated results on: %s [%s -- %s]" %(stamp, name, 'subtract reads'), 'yellow'))
    else:
        
        original_stdout = sys.stdout # Save a reference to the original standard output
        with open(reads_to_get, 'w') as f_out:
            sys.stdout = f_out # Change the standard output to the file we created.
            for line in open(bed_file_reduced):
                line=line.rstrip().split("\t")
                read=line[3].split("/")[0]
                print(read)

        sys.stdout = original_stdout
        f_out.close()
    
    ## convert bamtosam
    print("\t- Convert BAM to SAM format file only including subtracted reads...")
    sam_file = samtools_caller.bam2sam(name + '_reduced', bam_file, out_folder, ncpu, "-h -N " + reads_to_get, False, Debug)
    #                                sample, input_file, path_given, ncpu, options, sam2bam=False, Debug=False):

    ## convert sam to bam
    print("\t- Convert SAM to BAM format file only including subtracted reads...")
    bam_file_reduced = samtools_caller.bam2sam(name + '_reduced', sam_file, out_folder, ncpu, "", True, Debug)

    print("\t- Convert reduced BAM to BED format file...")
    bed_file_mapping = bedtools_caller.convert_bam2bed(name + '_reduced', bam_file_reduced, out_folder, pilfer=True, debug=Debug)
    
    ## New files
    pilfer_tmp = os.path.join(out_folder, name + ".pilfer.bed.tmp")
    pilfer_file = os.path.join(out_folder, name + ".pilfer.bed")
    
    ## Annotate reads using gold piRNA
    print("\t- Annotate reads in BAM using gold piRNA information provided...")
    
    ## generate paste filter tmp file
    filename_stamp = out_folder + '/.annotate_sam_success'
    if os.path.isfile(filename_stamp):
        stamp = HCGB_time.read_time_stamp(filename_stamp)
        print (colored("\tA previous command generated results on: %s [%s -- %s]" %(stamp, name, 'annotate sam'), 'yellow'))
        
        path_given = HCGB_files.create_folder(os.path.join(os.path.abspath(out_folder), "split_sam"))
        list_sam_parsed = HCGB_main.get_fullpath_list(path_given, Debug)
   
        ## remove non-desired files
        list_sam_parsed = [s for s in list_sam_parsed if not '.parsed' not in s]
        list_sam_parsed = [s for s in list_sam_parsed if '.split_file_success' not in s]

    else:
        list_sam_parsed = annotate_sam_call(sam_file, annot_info['piRBase']['gold_piRNA'], ncpu, out_folder, Debug)
        ## print time stamp
        HCGB_time.print_time_stamp(filename_stamp)
    
    ## merge annotated sam
    parsed_sam = os.path.join(out_folder, name + '_reduced-annotated.sam')
    list_sam_parsed = sorted(list_sam_parsed, key=lambda x: int("".join([i for i in x if i.isdigit()])))

    if Debug:
        HCGB_aes.debug_message("List of sam files parsed:")
        print(list_sam_parsed)

    ## add annotated reads into asingle file
    parsed_sam_file = open(parsed_sam, "w")
    for l in list_sam_parsed:
        for row in open(l, 'r'):
            parsed_sam_file.write(row)

    parsed_sam_file.close()
    
    #####
    print("\t- Create PILFER format file...")
    filename_stamp = out_folder + '/.convert_bam2pilfer_success'
    if os.path.isfile(filename_stamp):
        if (HCGB_files.is_non_zero_file(pilfer_tmp)):
            print (colored("\tA previous command generated results on: %s [%s -- %s]" %(stamp, name, 'bam2pilfer'), 'yellow'))
            return(pilfer_file)

    ## generate paste filter tmp file
    code_exec = merge_sam_bed(bed_file_mapping, parsed_sam, pilfer_tmp, Debug)
    if not code_exec:
        print ("** Some error ocurred while merging annotated SAM and BED file")
        exit()

    ## create bed file summarized:
    ## cat Aligned.sortedByCoord.out.sam.pilfer.bed.tmp | bedtools groupby -g 1,2,3,4,5 -c 4 -o count 

    bedtools_exe = set_config.get_exe("bedtools", Debug)
    cmd_bedtools = "%s sort -chrThenSizeA -i %s | sort -k 4 | %s groupby -o count -g 1,2,3,4,5 -c 4 | awk -v \"OFS=\t\" \'{print $1, $2, $3, $4, $6, $5}\' > %s " %(bedtools_exe, pilfer_tmp, bedtools_exe,  pilfer_file)
    bed_code = HCGB_sys.system_call(cmd_bedtools, False, True)
    if not bed_code:
        print("** Some error occurred while generating PILFER input file")
        exit()

    ## remove tmp files
    #os.remove(pilfer_tmp)
    #os.remove(sam_file_out)

    HCGB_time.print_time_stamp(filename_stamp)
    return(pilfer_file)

################################
def main():
    ## this code runs when call as a single script
    parser=argparse.ArgumentParser(description='''Create piRNA analysis using PILFER''');
    
    parser.add_argument('--input', '-i', help='Input BAM file', required=True);

    parser.add_argument('--name', '-n',
                        help='Name of the sample. Default: use filename provided.', default="");

    parser.add_argument('--path_given', '-p',
                        help='Name of the path to store results. Default: use filename provided.', default="");

    parser.add_argument('--database', 
                        help='Absolute path for database containing piRNA subfolder.', required=True);    

    parser.add_argument("-t", "--threads", type=int, help="Number of CPUs to use [Default: 2].", default=2)

    parser.add_argument("-s", "--species", help="Species abbreviation. [Default: hsa].", default="hsa")


    args=parser.parse_args();

    
    args.path_given = HCGB_files.create_folder(args.path_given)
    
    args.database = os.path.abspath(args.database)
    args.input = os.path.abspath(args.input)
    
    ## retrieved piRNA information
    annot_info = database.piRNA_info(args.database, args.species, False)
    
    ### 
    process_call(args.input, args.path_given, args.name, annot_info, args.threads, False)
    

################################
if __name__== "__main__":
    main()



#!/usr/bin/env python3
import setuptools
import glob
import HCGB.config.setup_module as setup_module

long_description_text = ""
with open("README.md", "r") as fh:
    long_description_text = fh.read()

setuptools.setup(
    name="XICRA",
    version=setup_module.get_version("./VERSION"),

    scripts=glob.glob('main/*'),
    author="Jose F. Sanchez-Herrero",

    author_email="jfbioinformatics@gmail.com",
    description="Small RNAseq pipeline for paired-end reads",

    long_description_content_type="text/markdown",
    long_description=long_description_text,
    
    url="https://github.com/HCGB-IGTP/XICRA/",
    packages=setuptools.find_packages(),
    license='MIT License',

    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    install_requires=setup_module.get_require_modules("XICRA/config/python/python_requirement_summary.txt"),

)



#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez                                        ##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain        ##
##########################################################
"""
Retrieves files within ``other_tools`` directory and returns path to given script specified
"""
## useful imports
import os
from HCGB.functions import files_functions, main_functions

####################################################################
def R_scripts(script, Debug):
    """
    Lists files within ``other_tools/R`` directory and returns path to given script
    """
    RDir = os.path.dirname(os.path.realpath(__file__))
    list_R = main_functions.get_fullpath_list(RDir, Debug)
    
    dict_R = {}
    for f in list_R:
        name = os.path.splitext(os.path.basename(f))[0]
        if (name == script):
            return (f)
        


#!/usr/bin/env python3
############################################################
## Jose F. Sanchez                                        ##
## Copyright (C) 2019-2020 Lauro Sumoy Lab, IGTP, Spain   ##
############################################################
## useful imports
import time
import io
import os
import re
import sys
from sys import argv
import subprocess
import argparse
from termcolor import colored

from XICRA.config import set_config
import HCGB.functions.aesthetics_functions as HCGB_aes
import HCGB.functions.files_functions as HCGB_files
import HCGB.functions.system_call_functions as HCGB_sys
import HCGB.functions.time_functions as HCGB_time

###################################################
def bam2sam(sample, input_file, path_given, ncpu, options, sam2bam=False, Debug=False):
    """
    This functions calls samtools to generate a conversion from BAM to SAM format.
    """
    ## create sam_file
    outfile_tmp = HCGB_files.get_path_name(input_file, path_given, sample, debug=Debug)
    
    if sam2bam:
        outfile = outfile_tmp.split('.sam')[0] + '.bam'
    else:
        outfile = outfile_tmp.split('.bam')[0] + '.sam'
    
    if not path_given:
        path_given = os.path.dirname(bam_file)
    
    filename_stamp = path_given + '/.convert_bam2sam_success'
    if os.path.isfile(filename_stamp):
        if HCGB_files.is_non_zero_file(outfile):
            stamp = HCGB_time.read_time_stamp(filename_stamp)
            print (colored("\tA previous command generated results on: %s [%s -- %s]" %(stamp, sample, 'samtools view conversion'), 'yellow'))
            return (outfile)

    ## Create call 
    samtools_exe = set_config.get_exe("samtools", Debug)
    cmd_samtools = "%s view --threads %s -o %s" %(samtools_exe, ncpu, outfile)
    
    if (options):
        cmd_samtools = cmd_samtools + " " + str(options)
    
    if (sam2bam):
        cmd_samtools = cmd_samtools + " -b"
    
    cmd_samtools = cmd_samtools + " " + input_file
    samtools_code = HCGB_sys.system_call(cmd_samtools, False, True)

    if samtools_code:
        ## print time stamp
        HCGB_time.print_time_stamp(filename_stamp)
    
        return (outfile)




#!/usr/bin/env python3
############################################################
## Jose F. Sanchez                                        ##
## Copyright (C) 2019-2020 Lauro Sumoy Lab, IGTP, Spain   ##
############################################################
import os
import re
import sys
import pandas as pd
import csv

from HCGB import functions
import HCGB.functions.aesthetics_functions as HCGB_aes

####################
def generate_DE(dataframe_results, Debug, outfolder, type_analysis='miRNA'):
	"""Builds final expression matrices comparing all samples.
	
		Generates three .csv for each software used:
		miRNA_expression-soft_name_dup.csv: counts with duplicated reads for each sample
		miRNA_expression-soft_name.csv: final matrix, counts of each isomiR (with miRNA and variant info) without duplicated reads 
		miRNA_expression-soft_name_seq.csv: table with the miRTop identifier and the corresponding DNA sequence
		
		miRNA is the default analysis but other can be provided such as tRNA, piRNA, etc
		
		:param dataframe_results: dataframe with the paths of the outputs of each sample and software
		:param Debug: display complete log
		:param outfolder: output folder
		
	    :returns: None
	"""
	
	## get results dictionary for each software employed 
	soft_list = dataframe_results.soft.unique()
	
	dataframe_results = dataframe_results.reset_index()
	
	## debugging messages
	if Debug:
		print ("## Debug:")
		print ("soft_list")
		print (soft_list)

	## generate results
	for soft_name in soft_list:
		
		print ("\n+ Summarizing results for software: ", soft_name)
		
		# retrieve files in dictionary
		tmp_df = dataframe_results[ dataframe_results['soft'] == soft_name].set_index('name')
		dict_files = tmp_df['filename'].to_dict()

		if Debug:
			print ("## Debug:")
			print ("tmp_df")
			print (tmp_df)
			print ("dict_files")
			print (dict_files)

		## get data and discard duplicate UIDs if any
		(all_data, all_seqs) = generate_matrix(dict_files, soft_name.lower(), Debug, type_analysis=type_analysis)
		all_data_filtered, all_data_duplicated = discard_UID_duplicated(all_data, type_res=type_analysis)
			
		
		## dump data in folder provided
		csv_outfile = os.path.join(outfolder, type_analysis + '_expression-' + soft_name)
		all_data_filtered.to_csv(csv_outfile + ".csv", quoting=csv.QUOTE_NONNUMERIC)
		all_data_duplicated.to_csv(csv_outfile + '_dup.csv', quoting=csv.QUOTE_NONNUMERIC)
		all_seqs.to_csv(csv_outfile + '_seq.csv', quoting=csv.QUOTE_NONNUMERIC)

####################
def discard_UID_duplicated(df_data, type_res="miRNA"):
	"""
	"""
	
	## get data index
	df_data['ID'] = df_data.index
	new_data = df_data.filter(['ID'], axis=1)	

	# split ID (hsa-let-7a-2-3p&NA&qNkjr6Ov2) into miRNA, variant and UID
	tmp = new_data['ID'].str.split('&', expand = True)
	
	new_data[type_res]  = tmp[0]
	new_data['variant']  = tmp[1]
	new_data['UID']  = tmp[2]
	
	## count 
	count_groups = new_data.groupby('UID').count()
	## print to file?
	
	## get duplicated
	bigger1count = count_groups[ count_groups['ID'] > 1 ]
	## print counts to file?
	
	## get list of UIDs duplicate
	bigger1count_list = bigger1count.index.to_list()
	duplicates = new_data[new_data['UID'].isin(bigger1count_list)]
	## print duplicate to file?

	## get duplicated data
	duplicates_indes_list = duplicates.index.to_list()
	duplicates_expression = df_data[df_data.index.isin(duplicates_indes_list)]
	#duplicates_expression['UID'] = duplicates_expression['ID'].str.split('&', expand = True)[2]
	duplicates_expression = duplicates_expression.drop(['ID'], axis=1)
	duplicates_expression.index.name = "ID"
	
	## get clean data
	clean_data_expression = df_data[~df_data.index.isin(duplicates_indes_list)]
	#clean_data_expression['UID'] = clean_data_expression['ID'].str.split('&', expand = True)[2]
	clean_data_expression =	clean_data_expression.drop(['ID'], axis=1)
	clean_data_expression.index.name = "ID"
	## Fix this error
	## A value is trying to be set on a copy of a slice from a DataFrame.
	## Try using .loc[row_indexer,col_indexer] = value instead
	## 
	## See the caveats in the documentation: http://pandas.pydata.org/pandas-docs/stable/indexing.html#indexing-view-versus-copy
	##   clean_data_expression['UID'] = clean_data_expression['ID'].str.split('&', expand = True)[2]
	
	return (clean_data_expression, duplicates_expression)

####################
def generate_matrix(dict_files, soft_name, Debug, type_analysis="miRNA"):
	"""
	"""
	######################################################
	### For a list of files, generates a count matrix with a unique index id by merging:
	### information provided within each file: name, variant and UID by '&'
	### e.g. AlaAGC&3'-tRF&tRF-16-KSP185D
	### e.g. hsa-let-7a-2-3p&NA&qNkjr6Ov2
	### 
	### It returns a dataframe containing for each index generated
	### count values for each sample (sample_name) in columns
	###
	#########################################################
	all_data = pd.DataFrame()
	seq_all_data = pd.DataFrame()
	for sample, this_file in dict_files.items():
		
		new_data=pd.DataFrame()
		
		print ('+ Reading information from sample: ', sample)	
		
		## 
		if functions.files_functions.is_non_zero_file(this_file):
			data = pd.read_csv(this_file, sep='\t')
		else:
			print ('\t - Information not available for sample: ', sample)	
			continue
		##
		if (data.size == 0):
			print ('\t - Information not available for sample: ', sample)  
			continue

		####
		if type_analysis=="miRNA":
			
			## ------------------------------------------ ##
			## Create matrix for miRNA results
			## ------------------------------------------ ##
			
			## get info, generate unique name and merge for samples
			## header of tsv files: 
			## UID	Read	miRNA	Variant	iso_5p	iso_3p	iso_add3p	iso_snp	sRNAbench
	
			## add NA if any
			data['Variant'] = data['Variant'].fillna('Canonical')
			
			
			## some variants are more complex and are denoted by several variants separated by comma:
			## e.g. iso_3p:+3,iso_add3p:1
			
			## These variants be included in different orders generating erroneous duplicated hits later:
			## e.g.: 
				## "hsa-miR-383-3p&iso_3p:+3,iso_add3p:1 & iso-22-0JEVN3JBF"
				## "hsa-miR-383-3p&iso_add3p:1,iso_3p:+3 & iso-22-0JEVN3JBF"

				## "hsa-miR-9500 & iso_add3p:1,iso_snv,iso_3p:+3 & iso-22-DKDERUKIQ"
				## "hsa-miR-9500 & iso_snv,iso_add3p:1,iso_3p:+3 & iso-22-DKDERUKIQ"

			## let's sort several entries if any and avoid this artifact

			## get variants that contain several types and sort them
			#print(data[data['Variant'].str.contains(",")])
			for i, row in data.iterrows():
				if (',' in row.Variant):
					list_of_variants = row['Variant'].split(',')
					list_of_variants.sort()
					data.at[i,'Variant'] = ",".join(list_of_variants)
								
			## create unique_id merging miRNA & variants & UID
			data['unique_id'] = data.apply(lambda data: data['miRNA'] + '&' + data['Variant'] + '&' + data['UID'], axis=1)
	
			## parse according to software
			if (soft_name == 'srnabench'):
				## sRNAbench mirtop creates a column id with sRNAbench instead of sample name
				new_data = data.filter(['unique_id', 'sRNAbench'], axis=1)
				new_data = new_data.set_index('unique_id')
				new_data = new_data.rename(columns={'sRNAbench': sample})
	
			if (soft_name == 'optimir'):
				## OptimiR mirtop creates a column containing sample name and other tags (trim, joined, fastq...)
				regex=re.compile(sample + '.*')
				search_list = list(filter(regex.match, data.columns.values.tolist()))
				new_data = data.filter(['unique_id', search_list[0]], axis=1)
				new_data = new_data.set_index('unique_id')
				new_data = new_data.rename(columns={search_list[0]: sample})
	
			if (soft_name == 'miraligner'):
				## miraligner mirtop creates a column containing sample name and other tags (trim, joined, fastq...)
				new_data = data.filter(['unique_id', sample], axis=1)
				new_data = new_data.set_index('unique_id')
		
		####
		elif "tRF" in type_analysis: ## tRF-amb; tRF-exc, tRF

			if Debug:
				HCGB_aes.debug_message(type_analysis + " analysis: ")

			## ------------------------------------------ ##
			## Create matrix for tRNA results
			## ------------------------------------------ ##
			## UID	Read	tRNA	variant	ident	expression	soft\n'
			
			data['variant'].fillna('NA', inplace=True)
			
			## parse according to software
			if (soft_name == 'mintmap'):
				data['unique_id'] = data.apply(lambda data: data['tRNA'] + '&' + data['variant'] + '&' + data['UID'], axis=1)
				new_data = data.filter(['unique_id', 'expression'], axis=1)
				new_data = new_data.set_index('unique_id')
				new_data = new_data.rename(columns={'expression': sample})
			
			## TODO
			#else:
			#	data['unique_id'] = data.apply(lambda data: data['tRNA'] + '&' + data['variant'] + '&' + data['UID'], axis=1)
			#	new_data = data.filter(['unique_id', 'XXXX'], axis=1) ## TODO
			#	new_data = new_data.set_index('unique_id')			  ## TODO
			#	new_data = new_data.rename(columns={'XXXX': sample})  ## TODO
				
		else:
			print()
			## add new
		
		## sequence information
		seq_data = data.filter(['UID', 'Read'], axis=1)	
		seq_data = seq_data.set_index('UID')
        
        ## debugging messages
		if Debug:
			HCGB_aes.debug_message("Data for samples: " + sample)
			print (new_data)
			HCGB_aes.debug_message("Sequence data for samples: " + sample)
			print (seq_data)
            
		#seq_all_data = seq_all_data.append(seq_data, sort=True).drop_duplicates('Read')
		seq_all_data = pd.concat([seq_all_data, seq_data], axis=1, sort=True).drop_duplicates('Read')
        
		all_data = pd.concat([all_data, new_data], axis=1, sort=True)

	##
	## debugging messages
	if Debug:
		HCGB_aes.debug_message("Data for all samples: ")
		print (all_data)
		HCGB_aes.debug_message("Data for sequences all samples: ")
		print (seq_all_data)
	
	return (all_data, seq_all_data)	

######

######
def main():
    ## this code runs when call as a single script

    ## control if options provided or help
    if len(sys.argv) > 1:
        print ("")
    else:
        print ("Provide file containing sample,file")
        exit()        

    dictionary_info = functions.main_functions.file2dictionary(sys.argv[1], ",")
    print (dictionary_info)

    ## get data
    (all_data, all_seqs) = generate_matrix(dictionary_info, "miraligner", False)

    ## discard duplicate UIDs if any
    all_data_filtered, all_data_duplicated = discard_UID_duplicated(all_data)

    ## dump data in folder provided
    outfolder = "./"
    csv_outfile = os.path.join(outfolder, 'miRNA_expression')
    all_data_filtered.to_csv(csv_outfile + ".csv", quoting=csv.QUOTE_NONNUMERIC)
    all_data_duplicated.to_csv(csv_outfile + '_dup.csv', quoting=csv.QUOTE_NONNUMERIC)
    all_seqs.to_csv(csv_outfile + '_seq.csv', quoting=csv.QUOTE_NONNUMERIC)



        
######
if __name__== "__main__":
    main()
    

